# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
import os.path
import urllib
import logging
import traceback
from contextlib import closing

import png

from weasy.document import PNGDocument
from weasy.tests.test_draw import assert_pixels_equal


TEST_SUITE_VERSION = '20110323'

BASE_URL = 'http://test.csswg.org/suites/css2.1/{}/html4/'.format(
    TEST_SUITE_VERSION)

RESULTS_DIRECTORY = os.path.join(os.path.dirname(__file__),
                                 'test_results', 'w3')


def get_url(url):
    return closing(urllib.urlopen(BASE_URL + url))


def get_test_list():
    with get_url('reftest.list') as reflist:
        for line in reflist:
            # Remove comments
            line = line.split('#', 1)[0]
            if not line.strip():
                # Comment-only line
                continue
            parts = line.split()
            comparaison = parts[0]
            if comparaison == '==':
                equal = True
            elif comparaison == '!=':
                equal = False
            else:
                raise ValueError(line)
            test = parts[1]
            references = parts[2:]
            assert references, 'No reference'
            yield equal, test, references


def make_test_suite():
    rendered = set()

    def render(name):
        filename = os.path.join(RESULTS_DIRECTORY, name + '.png')
        if name not in rendered:
            PNGDocument.from_file(BASE_URL + name).write_to(filename)
            rendered.add(name)
        return filename

    def make_test(equal, test, references):
        def test_function():
            test_filename = render(test)

            reader = png.Reader(filename=test_filename)
            test_width, test_height, test_lines, test_meta = reader.read()
            test_lines = list(test_lines)

            for reference in references:
                ref_filename = render(reference)

                reader = png.Reader(filename=ref_filename)
                ref_width, ref_height, ref_lines, ref_meta = reader.read()
                ref_lines = list(ref_lines)

                if equal:
                    assert test_width == ref_width
                    assert test_height == ref_height
                    assert_pixels_equal(test, test_width, test_height,
                                        test_lines, ref_lines)
                else:
                    assert test_lines != ref_lines

        return test_function

    for equal, test, references in get_test_list():
        yield test, make_test(equal, test, references)


def main():
    if not os.path.isdir(RESULTS_DIRECTORY):
        os.makedirs(RESULTS_DIRECTORY)

    cssutils_logger = logging.getLogger('CSSUTILS')
    del cssutils_logger.handlers[:]
    cssutils_logger.addHandler(logging.NullHandler())

    logger = logging.getLogger('WEASYPRINT')
    del logger.handlers[:]
    logger.addHandler(logging.NullHandler())

    tests = list(make_test_suite())
    passed = 0
    failed = 0
    errors = 0
    try:
        for i, (name, test) in enumerate(tests, 1):
            print '### Test %i of %i: %s' % (i, len(tests), name),
            sys.stdout.flush()
            try:
                test()
            except AssertionError:
                print 'FAIL'
                failed += 1
            except Exception:
                print 'ERROR:'
                traceback.print_exc()
                errors += 1
            else:
                print 'PASS'
                passed += 1
    except KeyboardInterrupt:
        print
        print 'Passed: %i, failed: %i, errors: %i' % (passed, failed, errors)

if __name__ == '__main__':
    main()



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Normalize values as much as possible without rendering the document.

"""

import collections

from .properties import INITIAL_VALUES
from .values import get_single_keyword, get_keyword


# How many CSS pixels is one <unit>?
# http://www.w3.org/TR/CSS21/syndata.html#length-units
LENGTHS_TO_PIXELS = {
    'px': 1,
    'pt': 1. / 0.75,
    'pc': 16.,  # LENGTHS_TO_PIXELS['pt'] * 12
    'in': 96.,  # LENGTHS_TO_PIXELS['pt'] * 72
    'cm': 96. / 2.54,  # LENGTHS_TO_PIXELS['in'] / 2.54
    'mm': 96. / 25.4,  # LENGTHS_TO_PIXELS['in'] / 25.4
}

# Value in pixels of font-size for <absolute-size> keywords: 12pt (16px) for
# medium, and scaling factors given in CSS3 for others:
# http://www.w3.org/TR/css3-fonts/#font-size-prop
# This dict has to be ordered to implement 'smaller' and 'larger'
FONT_SIZE_KEYWORDS = collections.OrderedDict(
    # medium is 16px, others are a ratio of medium
    (name, 16. * a / b)
    for name, a, b in [
        ('xx-small', 3, 5),
        ('x-small', 3, 4),
        ('small', 8, 9),
        ('medium', 1, 1),
        ('large', 6, 5),
        ('x-large', 3, 2),
        ('xx-large', 2, 1),
    ]
)

# These are unspecified, other than 'thin' <='medium' <= 'thick'.
# Values are in pixels.
BORDER_WIDTH_KEYWORDS = {
    'thin': 1,
    'medium': 3,
    'thick': 5,
}

# http://www.w3.org/TR/CSS21/fonts.html#propdef-font-weight
FONT_WEIGHT_RELATIVE = dict(
    bolder={
        100: 400,
        200: 400,
        300: 400,
        400: 700,
        500: 700,
        600: 900,
        700: 900,
        800: 900,
        900: 900,
    },
    lighter={
        100: 100,
        200: 100,
        300: 100,
        400: 100,
        500: 100,
        600: 400,
        700: 400,
        800: 700,
        900: 700,
    },
)

# http://www.w3.org/TR/css3-page/#size
# name=(width in pixels, height in pixels)
PAGE_SIZES = dict(
    A5=(
        148 * LENGTHS_TO_PIXELS['mm'],
        210 * LENGTHS_TO_PIXELS['mm'],
    ),
    A4=(
        210 * LENGTHS_TO_PIXELS['mm'],
        297 * LENGTHS_TO_PIXELS['mm'],
    ),
    A3=(
        297 * LENGTHS_TO_PIXELS['mm'],
        420 * LENGTHS_TO_PIXELS['mm'],
    ),
    B5=(
        176 * LENGTHS_TO_PIXELS['mm'],
        250 * LENGTHS_TO_PIXELS['mm'],
    ),
    B4=(
        250 * LENGTHS_TO_PIXELS['mm'],
        353 * LENGTHS_TO_PIXELS['mm'],
    ),
    letter=(
        8.5 * LENGTHS_TO_PIXELS['in'],
        11 * LENGTHS_TO_PIXELS['in'],
    ),
    legal=(
        8.5 * LENGTHS_TO_PIXELS['in'],
        14 * LENGTHS_TO_PIXELS['in'],
    ),
    ledger=(
        11 * LENGTHS_TO_PIXELS['in'],
        17 * LENGTHS_TO_PIXELS['in'],
    ),
)


class Computer(object):
    """Things that compute are computers, right? Handle `computed values`.

    Some computed values depend on other computed values. This object allow
    to request them without worrying about which is computed first.

    :param element: The HTML element these style apply to
    :param pseudo_type: The type of pseudo-element, eg 'before', None
    :param specified: a :class:`StyleDict` of specified values. Should contain
                      values for all properties.
    :param computed: a :class:`StyleDict` of already known computed values.
                     Only contains some properties (or none).
    :param parent_values: a :class:`StyleDict` of computed values of the parent
                          element (should contain values for all properties),
                          or ``None`` if ``element`` is the root element.

    Once instanciated, this object will have completed the ``computed`` dict
    so that is has values for all properties.

    """
    def __init__(self, element, pseudo_type, specified, computed,
                 parent_style):
        self.element = element
        self.pseudo_type = pseudo_type
        self.specified = specified
        self.parent_style = parent_style
        self.computed = computed

        for name in INITIAL_VALUES:
            self.get_computed(name)

    def get_computed(self, name):
        """Return the computed value for the ``name`` property.

        Call a "computer" function as needed and populate the `computed` dict
        before return the value.

        """
        if name in self.computed:
            # Already computed
            return self.computed[name]

        value = self.specified[name]
        if name in self.COMPUTER_FUNCTIONS:
            value = self.COMPUTER_FUNCTIONS[name](self, name, value)
        # else: same as specified

        assert value is not None
        self.computed[name] = value
        return value

    # Maps property names to functions returning the computed values
    COMPUTER_FUNCTIONS = {}

    @classmethod
    def register(cls, name):
        """Decorator registering a property ``name`` for a function."""
        name = name.replace('-', '_')
        def decorator(function):
            """Register the property ``name`` for ``function``."""
            cls.COMPUTER_FUNCTIONS[name] = function
            return function
        return decorator


# Let's be coherent, always use ``name`` as an argument even when it is useless
# pylint: disable=W0613

@Computer.register('background-color')
@Computer.register('border-top-color')
@Computer.register('border-right-color')
@Computer.register('border-bottom-color')
@Computer.register('border-left-color')
def other_color(computer, name, value):
    """Compute the ``*-color`` properties."""
    if value == 'currentColor':
        return computer.get_computed('color')
    else:
        # As specified
        return value


@Computer.register('color')
def color(computer, name, value):
    """Compute the ``color`` property."""
    if value == 'currentColor':
        if computer.parent_style is None:
            return INITIAL_VALUES['color']
        else:
            return computer.parent_style.color
    else:
        # As specified
        return value


@Computer.register('background-position')
def length_list(computer, name, values):
    """Compute the properties with a list of lengths."""
    return [length(computer, name, value) for value in values]


@Computer.register('top')
@Computer.register('right')
@Computer.register('left')
@Computer.register('bottom')
@Computer.register('margin-top')
@Computer.register('margin-right')
@Computer.register('margin-bottom')
@Computer.register('margin-left')
@Computer.register('height')
@Computer.register('width')
@Computer.register('letter-spacing')
@Computer.register('padding-top')
@Computer.register('padding-right')
@Computer.register('padding-bottom')
@Computer.register('padding-left')
@Computer.register('text-indent')
def length(computer, name, value):
    """Compute a length ``value``."""
    if getattr(value, 'type', 'other') == 'NUMBER' and value.value == 0:
        return 0

    if getattr(value, 'type', 'other') != 'DIMENSION':
        # No conversion needed.
        return value

    if value.dimension in LENGTHS_TO_PIXELS:
        # Convert absolute lengths to pixels
        factor = LENGTHS_TO_PIXELS[value.dimension]
    elif value.dimension in ('em', 'ex'):
        factor = computer.get_computed('font_size')

    if value.dimension == 'ex':
        factor *= 0.5

    return value.value * factor


@Computer.register('border-spacing')
def border_spacing(computer, name, value):
    value = tuple(length_list(computer, name, value))
    if len(value) == 1:
        return value * 2  # (v,) => (v, v)
    else:
        return value


@Computer.register('border-top-width')
@Computer.register('border-right-width')
@Computer.register('border-left-width')
@Computer.register('border-bottom-width')
def border_width(computer, name, value):
    """Compute the ``border-*-width`` properties."""
    style = computer.get_computed(name.replace('width', 'style'))
    if style in ('none', 'hidden'):
        return 0

    if value in BORDER_WIDTH_KEYWORDS:
        return BORDER_WIDTH_KEYWORDS[value]

    return length(computer, name, value)


@Computer.register('content')
def content(computer, name, values):
    """Compute the ``content`` property."""
    if computer.pseudo_type in ('before', 'after'):
        if values in ('normal', 'none'):
            return 'none'
        else:
            return [compute_content_value(computer, value) for value in values]
    else:
        # CSS 2.1 says it computes to 'normal' for elements, but does not say
        # anything for pseudo-elements other than :before and :after
        # (ie. :first-line and :first-letter)
        # Assume the same as elements.
        return 'normal'


def compute_content_value(computer, value):
    """Compute a content ``value``."""
    type_, content = value
    if type_ == 'attr':
        return ('STRING', computer.element.get(content, ''))
    else:
        return value


@Computer.register('display')
def display(computer, name, value):
    """Compute the ``display`` property.

    See http://www.w3.org/TR/CSS21/visuren.html#dis-pos-flo

    """
    float_ = computer.specified.float
    position = computer.specified.position
    if position in ('absolute', 'fixed') or float_ != 'none' or \
            computer.parent_style is None:
        if value == 'inline-table':
            return'table'
        elif value in ('inline', 'table-row-group', 'table-column',
                       'table-column-group', 'table-header-group',
                       'table-footer-group', 'table-row', 'table-cell',
                       'table-caption', 'inline-block'):
            return 'block'
    return value


@Computer.register('float')
def compute_float(computer, name, value):
    """Compute the ``float`` property.

    See http://www.w3.org/TR/CSS21/visuren.html#dis-pos-flo

    """
    position = computer.specified.position
    if position in ('absolute', 'fixed'):
        return 'none'
    else:
        return value


@Computer.register('font-size')
def font_size(computer, name, value):
    """Compute the ``font-size`` property."""
    if value in FONT_SIZE_KEYWORDS:
        return FONT_SIZE_KEYWORDS[value]

    if computer.parent_style is not None:
        parent_font_size = computer.parent_style.font_size
    else:
        # root element, no parent
        # Initial is 'medium', it’s a keyword.
        parent_font_size = FONT_SIZE_KEYWORDS[INITIAL_VALUES['font_size']]

    if value.type == 'DIMENSION':
        if value.dimension == 'px':
            factor = 1
        elif value.dimension == 'em':
            factor = parent_font_size
        elif value.dimension == 'ex':
            # TODO: find a better way to measure ex, see
            # http://www.w3.org/TR/CSS21/syndata.html#length-units
            factor = parent_font_size * 0.5
        elif value.dimension in LENGTHS_TO_PIXELS:
            factor = LENGTHS_TO_PIXELS[value.dimension]
    elif value.type == 'PERCENTAGE':
        factor = parent_font_size / 100.
    elif value.type == 'NUMBER' and value.value == 0:
        return 0

    # Raise if `factor` is not defined. It should be, because of validation.
    return value.value * factor


@Computer.register('font-weight')
def font_weight(computer, name, value):
    """Compute the ``font-weight`` property."""
    if value == 'normal':
        return 400
    elif value == 'bold':
        return 700
    elif value in ('bolder', 'lighter'):
        if computer.parent_style is not None:
            parent_value = computer.parent_style.font_weight
        else:
            initial = get_single_keyword(INITIAL_VALUES['font_weight'])
            assert initial == 'normal'
            parent_value = 400
        # Use a string here as StyleDict.__setattr__ turns integers into pixel
        # lengths. This is a number without unit.
        return FONT_WEIGHT_RELATIVE[value][parent_value]
    else:
        return value


@Computer.register('line-height')
def line_height(computer, name, value):
    """Compute the ``line-height`` property."""
    if value == 'normal':
        # a "reasonable" value
        # http://www.w3.org/TR/CSS21/visudet.html#line-height
        # TODO: use font metadata?
        factor = 1.2
    elif value.type == 'NUMBER':
        factor = value.value
    elif value.type == 'PERCENTAGE':
        factor = value.value / 100.
    elif value.type == 'DIMENSION':
        return length(computer, name, value)
    font_size_value = computer.get_computed('font_size')
    # Raise if `factor` is not defined. It should be, because of validation.
    return factor * font_size_value


@Computer.register('size')
def size(computer, name, values):
    """Compute the ``size`` property.

    See CSS3 Paged Media.

    """
    if computer.element != '@page':
        return 'none'

    keywords = map(get_keyword, values)
    values = length_list(computer, name, values)

    if keywords == ['auto']:
        keywords = ['A4']  # Chosen by the UA. (That’s me!)

    if isinstance(values[0], (int, float)):
        if len(values) == 2:
            assert isinstance(values[1], (int, float))
            return values
        else:
            # square page
            value, = values
            return value, value
    else:
        orientation = None
        size_value = None
        for i, keyword in enumerate(keywords):
            if keyword in ('portrait', 'landscape'):
                orientation = keyword
            elif keyword in PAGE_SIZES:
                size_value = keyword
            else:
                raise ValueError("Illegal value for 'size': %r" % values[i])
        if size_value is None:
            size_value = 'A4'
        width, height = PAGE_SIZES[size_value]
        if (orientation == 'portrait' and width > height) or \
               (orientation == 'landscape' and height > width):
            width, height = height, width
        return width, height


@Computer.register('text-align')
def text_align(computer, name, value):
    """Compute the ``text-align`` property."""
    if value == 'start':
        if computer.get_computed('direction') == 'rtl':
            return 'right'
        else:
            return 'left'
    else:
        return value


@Computer.register('vertical_align')
def vertical_align(computer, name, value):
    """Compute the ``word-spacing`` property."""
    # Use +/- half an em for super and sub, same as Pango.
    # (See the SUPERSUB_RISE constant in pango-markup.c)
    if value == 'super':
        return computer.get_computed('font_size') * 0.5
    if value == 'sub':
        return computer.get_computed('font_size') * -0.5
    if getattr(value, 'type', 'other') == 'PERCENTAGE':
        return computer.get_computed('line_height') * value.value / 100.
    return length(computer, name, value)


@Computer.register('word-spacing')
def word_spacing(computer, name, value):
    """Compute the ``word-spacing`` property."""
    if value == 'normal':
        return 0

    return length(computer, name, value)



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Various utils.

"""

import urllib
from urlparse import urljoin, urlparse

from cssutils.helper import path2url

from . import VERSION


def get_url_attribute(element, key):
    """Get the URL corresponding to the ``key`` attribute of ``element``.

    The retrieved URL is absolute, even if the URL in the element is relative.

    """
    attr_value = element.get(key)
    if attr_value is None:
        return None
    return urljoin(element.base_url, attr_value.strip())


def ensure_url(string):
    """Get a ``scheme://path`` URL from ``string``.

    If ``string`` looks like an URL, return it unchanged. Otherwise assume a
    filename and convert it to a ``file://`` URL.

    """
    return string if urlparse(string).scheme else path2url(string)


class URLopener(urllib.FancyURLopener):
    # User-Agent
    version = 'WeasyPrint/%s http://weasyprint.org/' % VERSION


def urlopen(url):
    """Fetch an URL and return ``(file_like, mime_type, charset)``.

    It is the caller’s responsability to call ``file_like.close()``.
    """
    file_like = URLopener().open(url)
    info = file_like.info()
    if hasattr(info, 'get_content_type'):
        # Python 3
        mime_type = info.get_content_type()
    else:
        # Python 2
        mime_type = info.gettype()
    if hasattr(info, 'get_param'):
        # Python 3
        charset = info.get_param('charset')
    else:
        # Python 2
        charset = info.getparam('charset')
    return file_like.fp, mime_type, charset


def urllib_fetcher(url):
    """URL fetcher for cssutils.

    This fetcher is based on urllib instead of urllib2, since urllib has
    support for the "data" URL scheme.

    """
    file_like, mime_type, charset = urlopen(url)
    if mime_type != 'text/css':
        # TODO: add a warning
        return None
    content = file_like.read()
    file_like.close()
    return charset, content



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
WeasyPrint testing suite.

"""

import os.path
import logging

from cssutils import parseFile

from ..document import PNGDocument


logging.getLogger('WEASYPRINT').addHandler(logging.StreamHandler())


# TODO: find a way to not depend on a specific font
FONTS = u"Liberation Sans, Arial"
TEST_USER_AGENT_STYLESHEETS = (
    parseFile(os.path.join(
        os.path.dirname(__file__), '..', 'css', 'tests_ua.css'
    )),
)


class TestPNGDocument(PNGDocument):
    """Like PNGDocument, but with a different user-agent stylesheet.

    This stylesheet is shorter, which makes tests faster.

    """
    def __init__(self, dom, user_stylesheets=None,
                 user_agent_stylesheets=TEST_USER_AGENT_STYLESHEETS):
        super(TestPNGDocument, self).__init__(
            dom, user_stylesheets, user_agent_stylesheets)


def resource_filename(basename):
    """Return the absolute path of the resource called ``basename``."""
    return os.path.join(os.path.dirname(__file__), 'resources', basename)



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Output document classes for various formats.

"""

import os.path
import math
import logging

from cssutils import CSSParser
import lxml.html
import cairo

from .css import get_all_computed_styles
from .css.computed_values import LENGTHS_TO_PIXELS
from .formatting_structure.build import build_formatting_structure
from .layout import layout
from . import draw
from . import utils
from . import images


logger = logging.getLogger('CSSUTILS')
level = logger.level
logger.setLevel('ERROR')
# This is a one-element tuple.
DEFAULT_USER_AGENT_STYLESHEETS = (
    CSSParser(parseComments=False, validate=False).parseFile(
        os.path.join(os.path.dirname(__file__), 'css', 'html5_ua.css')
    ),
)
logger.setLevel(level)
del logger, level


class Document(object):
    """Abstract output document."""
    def __init__(self, dom, user_stylesheets=None,
                 user_agent_stylesheets=DEFAULT_USER_AGENT_STYLESHEETS):
        assert getattr(dom, 'tag', None) == 'html', (
            'HTML document expected, got %r.' % (dom,))

        #: lxml HtmlElement object
        self.dom = dom

        # Go through the property setter which calls ensure_url()
        self.base_url = self.base_url

        self.css_parser = CSSParser(parseComments=False, validate=False)
        self.css_parser.setFetcher(utils.urllib_fetcher)

        self.user_stylesheets = user_stylesheets or []
        self.user_agent_stylesheets = user_agent_stylesheets or []

        self._computed_styles = None
        self._formatting_structure = None
        self._pages = None
        self._image_cache = {}

    @property
    def base_url(self):
        """The URL of the document, used for relative URLs it contains.

        If set to something that does not look like a URL, the value is
        assumed to be a filename and is converted to a file:// URL.
        If that filename is relative, it is interpreted from the current
        directory.

        """
        return self.dom.getroottree().docinfo.URL

    @base_url.setter
    def base_url(self, value):
        if value:
            value = utils.ensure_url(value)
        self.dom.getroottree().docinfo.URL = value

    @classmethod
    def from_string(cls, source, encoding=None, **kwargs):
        """Make a document from an HTML string."""
        parser = lxml.html.HTMLParser(encoding=encoding)
        dom = lxml.html.document_fromstring(source, parser=parser)
        return cls(dom, **kwargs)

    @classmethod
    def from_file(cls, file_or_filename_or_url, encoding=None, **kwargs):
        """
        Make a document from a filename or open file object.
        """
        parser = lxml.html.HTMLParser(encoding=encoding)
        dom = lxml.html.parse(file_or_filename_or_url, parser=parser).getroot()
        return cls(dom, **kwargs)

    def style_for(self, element, pseudo_type=None):
        """
        Convenience method to get the computed styles for an element.
        """
        return self.computed_styles.get((element, pseudo_type))

    @property
    def computed_styles(self):
        """
        dict of (element, pseudo_element_type) -> StyleDict
        StyleDict: a dict of property_name -> PropertyValue,
                   also with attribute access
        """
        if self._computed_styles is None:
            self._computed_styles = get_all_computed_styles(
                self,
                user_stylesheets=self.user_stylesheets,
                ua_stylesheets=self.user_agent_stylesheets,
                medium='print')
        return self._computed_styles

    @property
    def formatting_structure(self):
        """
        The Box object for the root element. Represents the tree of all boxes.
        """
        if self._formatting_structure is None:
            self._formatting_structure = build_formatting_structure(self)
        return self._formatting_structure

    @property
    def pages(self):
        """
        List of layed-out pages with an absolute size and postition
        for every box.
        """
        if self._pages is None:
            # "Linearize" code flow
            _ = self.computed_styles
            _ = self.formatting_structure
            # Actual work
            self._pages = layout(self)
        return self._pages

    def get_image_surface_from_uri(self, uri):
        missing = object()
        surface = self._image_cache.get(uri, missing)
        if surface is missing:
            surface = images.get_image_surface_from_uri(uri)
            self._image_cache[uri] = surface
        return surface


class PNGDocument(Document):
    """PNG output document."""
    def __init__(self, dom, *args, **kwargs):
        super(PNGDocument, self).__init__(dom, *args, **kwargs)
        self.surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1, 1)

    def draw_page(self, page):
        """Draw a single page and return an ImageSurface."""
        width = int(math.ceil(page.outer_width))
        height = int(math.ceil(page.outer_height))
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        context = draw.CairoContext(surface)
        draw.draw_page(self, page, context)
        self.surface.finish()
        return width, height, surface

    def write_page_to(self, page_index, target):
        """Write a single page as PNG into a file-like or filename `target`."""
        _width, _height, surface = self.draw_page(self.pages[page_index])
        surface.write_to_png(target)

    def write_to(self, target):
        """Write all pages as PNG into a file-like or filename `target`.

        Pages are layed out vertically each above the next and centered
        horizontally.
        """
        pages = [self.draw_page(page) for page in self.pages]
        total_height = sum(height for width, height, surface in pages)
        max_width = max(width for width, height, surface in pages)
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32,
            max_width, total_height)
        context = draw.CairoContext(surface)

        position_y = 0
        for width, height, page_surface in pages:
            position_x = (max_width - width) // 2
            context.set_source_surface(page_surface, position_x, position_y)
            context.paint()
            position_y += height

        surface.write_to_png(target)


class PDFDocument(Document):
    """PDF output document."""
    def __init__(self, dom, *args, **kwargs):
        super(PDFDocument, self).__init__(dom, *args, **kwargs)
        # Use a dummy page size initially
        self.surface = cairo.PDFSurface(None, 1, 1)

    def write_to(self, target):
        """
        Write the whole document as PDF into a file-like or filename `target`.
        """
        # The actual page size is set for each page.
        surface = cairo.PDFSurface(target, 1, 1)

        px_to_pt = 1 / LENGTHS_TO_PIXELS['pt']
        for page in self.pages:
            # Actual page size is here. May be different between pages.
            surface.set_size(
                page.outer_width * px_to_pt,
                page.outer_height * px_to_pt)
            context = draw.CairoContext(surface)
            context.scale(px_to_pt, px_to_pt)
            draw.draw_page(self, page, context)
            surface.show_page()

        surface.finish()



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Module drawing documents.

"""

from __future__ import division
import contextlib

import cairo

from .formatting_structure import boxes
from .css.values import get_percentage_value


class CairoContext(cairo.Context):
    """A ``cairo.Context`` with a few more helper methods."""

    def set_source_colorvalue(self, color):
        """Set the source pattern from a ``cssutils.ColorValue`` object."""
        self.set_source_rgba(
            color.red / 255., color.green / 255., color.blue / 255.,
            color.alpha)

    @contextlib.contextmanager
    def stacked(self):
        """Save and restore the context when used with the ``with`` keyword."""
        self.save()
        try:
            yield
        finally:
            self.restore()


def draw_page(document, page, context):
    """Draw the given PageBox to a Cairo context.

    The context should be scaled so that lengths are in CSS pixels.

    """
    draw_page_background(document, context, page)
    draw_border(context, page)
    draw_box(document, context, page.root_box)


def draw_box(document, context, box):
    """Draw a ``box`` on ``context``."""
    if box.style.visibility == 'visible':
        if has_background(box):
            draw_background(document, context, box)

        draw_border(context, box)

        marker_box = getattr(box, 'outside_list_marker', None)
        if marker_box:
            draw_box(document, context, marker_box)

        if isinstance(box, boxes.TextBox):
            draw_text(context, box)
            return

        if isinstance(box, boxes.ReplacedBox):
            draw_replacedbox(context, box)

    if isinstance(box, boxes.TableBox):
        for child in box.column_groups:
            draw_box(document, context, child)

    if isinstance(box, boxes.ParentBox):
        for child in box.children:
            draw_box(document, context, child)



def has_background(box):
    """Return whether the given box has any background."""
    return box.style.background_color.alpha > 0 or \
        box.style.background_image != 'none'


def draw_page_background(document, context, page):
    """Draw the backgrounds for the page box (from @page style) and for the
    page area (from the root element).

    If the root element is "html" and has no background, the page area’s
    background is taken from its "body" child.

    In both cases the background position is the same as if it was drawn on
    the element.

    See http://www.w3.org/TR/CSS21/colors.html#background

    """
    # TODO: this one should have its origin at (0, 0), not the border box
    # of the page.
    # TODO: more tests for this, see
    # http://www.w3.org/TR/css3-page/#page-properties
    draw_background(document, context, page, clip=False)
    if has_background(page.root_box):
        draw_background(document, context, page.root_box, clip=False)
    elif page.root_box.element_tag.lower() == 'html':
        for child in page.root_box.children:
            if child.element_tag.lower() == 'body':
                # This must be drawn now, before anything on the root element.
                draw_background(document, context, child, clip=False)


def draw_background(document, context, box, clip=True):
    """Draw the box background color and image to a ``cairo.Context``."""
    if getattr(box, 'background_drawn', False):
        return

    box.background_drawn = True

    if not has_background(box):
        return

    with context.stacked():
        bg_x = box.border_box_x()
        bg_y = box.border_box_y()
        bg_width = box.border_width()
        bg_height = box.border_height()

        bg_attachement = box.style.background_attachment
        if bg_attachement == 'fixed':
            # There should not be any clip yet
            x1, y1, x2, y2 = context.clip_extents()
            page_width = x2 - x1
            page_height = y2 - y1

        if clip:
            context.rectangle(bg_x, bg_y, bg_width, bg_height)
            context.clip()

        # Background color
        bg_color = box.style.background_color
        if bg_color.alpha > 0:
            context.set_source_colorvalue(bg_color)
            context.paint()

        if bg_attachement == 'scroll':
            # Change coordinates to make the rest easier.
            context.translate(bg_x, bg_y)
        else:
            assert bg_attachement == 'fixed'
            bg_width = page_width
            bg_height = page_height

        # Background image
        bg_image = box.style.background_image
        if bg_image == 'none':
            return

        image = document.get_image_surface_from_uri(bg_image)
        if image is None:
            return

        surface, image_width, image_height = image

        bg_position = box.style.background_position
        bg_position_x, bg_position_y = absolute_background_position(
            bg_position, (bg_width, bg_height), (image_width, image_height))
        context.translate(bg_position_x, bg_position_y)

        bg_repeat = box.style.background_repeat
        if bg_repeat != 'repeat':
            # Get the current clip rectangle
            clip_x1, clip_y1, clip_x2, clip_y2 = context.clip_extents()
            clip_width = clip_x2 - clip_x1
            clip_height = clip_y2 - clip_y1

            if bg_repeat in ('no-repeat', 'repeat-x'):
                # Limit the drawn area vertically
                clip_y1 = 0  # because of the last context.translate()
                clip_height = image_height

            if bg_repeat in ('no-repeat', 'repeat-y'):
                # Limit the drawn area horizontally
                clip_x1 = 0  # because of the last context.translate()
                clip_width = image_width

            # Second clip for the background image
            context.rectangle(clip_x1, clip_y1, clip_width, clip_height)
            context.clip()

        context.set_source_surface(surface)
        context.get_source().set_extend(cairo.EXTEND_REPEAT)
        context.paint()


def absolute_background_position(css_values, bg_dimensions, image_dimensions):
    """Return the background's ``position_x, position_y`` in pixels.

    http://www.w3.org/TR/CSS21/colors.html#propdef-background-position

    :param css_values: a list of one or two cssutils Value objects.
    :param bg_dimensions: ``width, height`` of the background positionning area
    :param image_dimensions: ``width, height`` of the background image

    """
    values = list(css_values)

    if len(css_values) == 1:
        values.append('center')
    else:
        assert len(css_values) == 2

    if values[1] in ('left', 'right') or values[0] in ('top', 'bottom'):
        values.reverse()
    # Order is now [horizontal, vertical]

    kw_to_percentage = dict(top=0, left=0, center=50, bottom=100, right=100)

    for value, bg_dimension, image_dimension in zip(
            values, bg_dimensions, image_dimensions):
        percentage = kw_to_percentage.get(value, get_percentage_value(value))
        if percentage is not None:
            yield (bg_dimension - image_dimension) * percentage / 100.
        else:
            yield value


def get_rectangle_edges(x, y, width, height):
    """Return the 4 edges of a rectangle as a list.

    Edges are in clock-wise order, starting from the top.

    Each edge is returned as ``(start_point, end_point)`` and each point
    as ``(x, y)`` coordinates.

    """
    # In clock-wise order, starting on top left
    corners = [
        (x, y),
        (x + width, y),
        (x + width, y + height),
        (x, y + height)]
    # clock-wise order, starting on top right
    shifted_corners = corners[1:] + corners[:1]
    return zip(corners, shifted_corners)


def draw_border(context, box):
    """Draw the box border to a ``cairo.Context``."""
    if all(box.style['border_%s_width' % side] == 0
           for side in ['top', 'right', 'bottom', 'left']):
        # No border, return early.
        return

    for side, x_offset, y_offset, border_edge, padding_edge in zip(
        ['top', 'right', 'bottom', 'left'],
        [0, -1, 0, 1],
        [1, 0, -1, 0],
        get_rectangle_edges(
            box.border_box_x(), box.border_box_y(),
            box.border_width(), box.border_height(),
        ),
        get_rectangle_edges(
            box.padding_box_x(), box.padding_box_y(),
            box.padding_width(), box.padding_height(),
        ),
    ):
        width = box.style['border_%s_width' % side]
        if width == 0:
            continue
        color = box.style['border_%s_color' % side]
        if color.alpha == 0:
            continue
        style = box.style['border_%s_style' % side]
        with context.stacked():
            """
            Both edges form a trapezoid. This is the top one:

              +---------------+
               \             /
              =================
                 \         /
                  +-------+

            We clip on its outline on draw on the big line on the middle.
            """
            # I find that anti-aliasing does not look good on borders.
            context.set_antialias(cairo.ANTIALIAS_NONE)
            # TODO: implement other styles.
            if not style in ['dotted', 'dashed']:
                border_start, border_stop = border_edge
                padding_start, padding_stop = padding_edge
                # Move to one of the Trapezoid’s corner
                context.move_to(*border_start)
                for point in [border_stop, padding_stop,
                              padding_start, border_start]:
                    context.line_to(*point)
                context.clip()
            elif style == 'dotted':
                # TODO: find a way to make a real dotted border
                context.set_dash([width], 0)
            elif style == 'dashed':
                # TODO: find a way to make a real dashed border
                context.set_dash([4 * width], 0)
            (x1, y1), (x2, y2) = border_edge
            offset = width / 2
            x_offset *= offset
            y_offset *= offset
            x1 += x_offset
            x2 += x_offset
            y1 += y_offset
            y2 += y_offset
            context.move_to(x1, y1)
            context.line_to(x2, y2)
            context.set_source_colorvalue(color)
            context.set_line_width(width)
            context.stroke()


def draw_replacedbox(context, box):
    """Draw the given :class:`boxes.ReplacedBox` to a ``cairo.context``."""
    x, y = box.padding_box_x(), box.padding_box_y()
    width, height = box.width, box.height
    surface, intrinsic_width, intrinsic_height = box.replacement
    with context.stacked():
        context.translate(x, y)
        context.rectangle(0, 0, width, height)
        context.clip()
        scale_width = width / intrinsic_width
        scale_height = height / intrinsic_height
        context.scale(scale_width, scale_height)
        context.set_source_surface(surface)
        context.paint()

def draw_text(context, textbox):
    """Draw ``textbox`` to a ``cairo.Context`` from ``PangoCairo.Context``."""
    # Pango crashes with font-size: 0
    assert textbox.style.font_size

    context.move_to(textbox.position_x, textbox.position_y + textbox.baseline)
    context.set_source_colorvalue(textbox.style.color)
    textbox.show_line(context)
    values = textbox.style.text_decoration
    for value in values:
        if value == 'overline':
            draw_overline(context, textbox)
        elif value == 'underline':
            draw_underline(context, textbox)
        elif value == 'line-through':
            draw_line_through(context, textbox)


def draw_overline(context, textbox):
    """Draw overline of ``textbox`` to a ``cairo.Context``."""
    font_size = textbox.style.font_size
    position_y = textbox.baseline + textbox.position_y - (font_size * 0.15)
    draw_text_decoration(context, position_y, textbox)


def draw_underline(context, textbox):
    """Draw underline of ``textbox`` to a ``cairo.Context``."""
    font_size = textbox.style.font_size
    position_y = textbox.baseline + textbox.position_y + (font_size * 0.15)
    draw_text_decoration(context, position_y, textbox)


def draw_line_through(context, textbox):
    """Draw line-through of ``textbox`` to a ``cairo.Context``."""
    position_y = textbox.position_y + (textbox.height * 0.5)
    draw_text_decoration(context, position_y, textbox)


def draw_text_decoration(context, position_y, textbox):
    """Draw text-decoration of ``textbox`` to a ``cairo.Context``."""
    with context.stacked():
        context.set_source_colorvalue(textbox.style.color)
        context.set_line_width(1)  # TODO: make this proportional to font_size?
        context.move_to(textbox.position_x, position_y)
        context.rel_line_to(textbox.width, 0)
        context.stroke()



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Functions laying out the inline boxes.

"""

from __future__ import division

import cairo

from .markers import image_marker_layout
from .percentages import resolve_percentages, resolve_one_percentage
from ..text import TextFragment
from ..formatting_structure import boxes


def get_next_linebox(document, linebox, position_y, skip_stack,
                     containing_block, device_size):
    """Return ``(line, resume_at)``.

    ``line`` is a laid-out LineBox with as much content as possible that
    fits in the available width.

    :param linebox: a non-laid-out :class:`LineBox`
    :param position_y: vertical top position of the line box on the page
    :param skip_stack: ``None`` to start at the beginning of ``linebox``,
                       or a ``resume_at`` value to continue just after an
                       already laid-out line.
    :param containing_block: Containing block of the line box:
                             a :class:`BlockContainerBox`
    :param device_size: ``(width, height)`` of the current page.

    """
    position_x = linebox.position_x
    max_x = position_x + containing_block.width
    if skip_stack is None:
        # text-indent only at the start of the first line
        # Other percentages (margins, width, ...) do not apply.
        resolve_one_percentage(linebox, 'text_indent', containing_block.width)
        position_x += linebox.text_indent

    skip_stack = skip_first_whitespace(linebox, skip_stack)
    if skip_stack == 'continue':
        return None, None

    resolve_percentages(linebox, containing_block)
    line, resume_at, preserved_line_break = split_inline_box(
        document, linebox, position_x, max_x, skip_stack, containing_block,
        device_size)

    remove_last_whitespace(document, line)
    offset_x = text_align(line, containing_block)

    bottom, top = inline_box_verticality(line, baseline_y=0)
    if bottom is None:
        # No children at all
        line.position_y = position_y
        offset_y = 0
        if preserved_line_break:
            # Only the strut.
            line.baseline = line.margin_top
            line.height += line.margin_top + line.margin_bottom
        else:
            line.height = 0
            line.baseline = 0
    else:
        assert top is not None
        line.baseline = -top
        line.position_y = top
        line.height = bottom - top
        offset_y = position_y - top
    line.margin_top = 0
    line.margin_bottom = 0
    if offset_x != 0 or offset_y != 0:
        # This also translates children
        line.translate(offset_x, offset_y)
    return line, resume_at


def skip_first_whitespace(box, skip_stack):
    """Return the ``skip_stack`` to start just after the remove spaces
    at the beginning of the line.

    See http://www.w3.org/TR/CSS21/text.html#white-space-model
    """
    if skip_stack is None:
        index = 0
        next_skip_stack = None
    else:
        index, next_skip_stack = skip_stack

    if isinstance(box, boxes.TextBox):
        assert next_skip_stack is None
        white_space = box.style.white_space
        length = len(box.text)
        if index == length:
            # Starting a the end of the TextBox, no text to see: Continue
            return 'continue'
        if white_space in ('normal', 'nowrap', 'pre-line'):
            while index < length and box.text[index] == ' ':
                index += 1
        return index, None

    if isinstance(box, (boxes.LineBox, boxes.InlineBox)):
        if index == 0 and not box.children:
            return None
        result = skip_first_whitespace(box.children[index], next_skip_stack)
        if result == 'continue':
            index += 1
            if index >= len(box.children):
                return 'continue'
            result = skip_first_whitespace(box.children[index], None)
        return index, result

    assert skip_stack is None, 'unexpected skip inside %s' % box
    return None


def remove_last_whitespace(document, box):
    """Remove in place space characters at the end of a line.

    This also reduces the width of the inline parents of the modified text.

    """
    ancestors = []
    while isinstance(box, (boxes.LineBox, boxes.InlineBox)):
        ancestors.append(box)
        if not box.children:
            return
        box = box.children[-1]
    if not (isinstance(box, boxes.TextBox) and
            box.style.white_space in ('normal', 'nowrap', 'pre-line')):
        return
    new_text = box.text.rstrip(' ')
    if new_text:
        if len(new_text) == len(box.text):
            return
        new_box, resume, _ = split_text_box(document, box, box.width * 2, 0)
        assert new_box is not None
        assert resume is None
        space_width = box.width - new_box.width
        box.width = new_box.width
        box.show_line = new_box.show_line
    else:
        space_width = box.width
        box.width = 0
        box.show_line = lambda x: x  # No-op
    box.text = new_text
    for ancestor in ancestors:
        ancestor.width -= space_width

    # TODO: All tabs (U+0009) are rendered as a horizontal shift that
    # lines up the start edge of the next glyph with the next tab stop.
    # Tab stops occur at points that are multiples of 8 times the width
    # of a space (U+0020) rendered in the block's font from the block's
    # starting content edge.

    # TODO: If spaces (U+0020) or tabs (U+0009) at the end of a line have
    # 'white-space' set to 'pre-wrap', UAs may visually collapse them.


def inline_replaced_box_layout(box, containing_block, device_size):
    """Lay out an inline :class:`boxes.ReplacedBox` ``box``."""
    assert isinstance(box, boxes.ReplacedBox)

    # Compute width:
    # http://www.w3.org/TR/CSS21/visudet.html#inline-replaced-width
    if box.margin_left == 'auto':
        box.margin_left = 0
    if box.margin_right == 'auto':
        box.margin_right = 0
    replaced_box_width(box, device_size)

    # Compute height
    # http://www.w3.org/TR/CSS21/visudet.html#inline-replaced-height
    if box.margin_top == 'auto':
        box.margin_top = 0
    if box.margin_bottom == 'auto':
        box.margin_bottom = 0
    replaced_box_height(box, device_size)


def replaced_box_width(box, device_size):
    """
    Compute and set the used width for replaced boxes (inline- or block-level)
    """
    _surface, intrinsic_width, intrinsic_height = box.replacement
    intrinsic_ratio = intrinsic_width / intrinsic_height

    if box.height == 'auto' and box.width == 'auto':
        if intrinsic_width is not None:
            box.width = intrinsic_width
        elif intrinsic_height is not None and intrinsic_ratio is not None:
            box.width = intrinsic_ratio * intrinsic_height
        elif box.height != 'auto' and intrinsic_ratio is not None:
            box.width = intrinsic_ratio * box.height
        elif intrinsic_ratio is not None:
            pass
            # TODO: Intrinsic ratio only: undefined in CSS 2.1.
            # " It is suggested that, if the containing block's width does not
            #   itself depend on the replaced element's width, then the used
            #   value of 'width' is calculated from the constraint equation
            #   used for block-level, non-replaced elements in normal flow. "

    # Still no value
    if box.width == 'auto':
        if intrinsic_width is not None:
            box.width = intrinsic_width
        else:
            # Then the used value of 'width' becomes 300px. If 300px is too
            # wide to fit the device, UAs should use the width of the largest
            # rectangle that has a 2:1 ratio and fits the device instead.
            device_width, _device_height = device_size
            box.width = min(300, device_width)


def replaced_box_height(box, device_size):
    """
    Compute and set the used height for replaced boxes (inline- or block-level)
    """
    _surface, intrinsic_width, intrinsic_height = box.replacement
    intrinsic_ratio = intrinsic_width / intrinsic_height

    if box.height == 'auto' and box.width == 'auto':
        if intrinsic_height is not None:
            box.height = intrinsic_height
    elif intrinsic_ratio is not None and box.height == 'auto':
        box.height = box.width / intrinsic_ratio
    elif box.height == 'auto' and intrinsic_height is not None:
        box.height = intrinsic_height
    elif box.height == 'auto':
        device_width, _device_height = device_size
        box.height = min(150, device_width / 2)


def atomic_box(box, containing_block, device_size):
    """Compute the width and the height of the atomic ``box``."""
    if isinstance(box, boxes.ReplacedBox):
        if getattr(box, 'is_list_marker', False):
            image_marker_layout(box)
        else:
            inline_replaced_box_layout(box, containing_block, device_size)
    else:
        raise TypeError('Layout for %s not handled yet' % type(box).__name__)
    return box


def split_inline_level(document, box, position_x, max_x, skip_stack,
                       containing_block, device_size):
    """Fit as much content as possible from an inline-level box in a width.

    Return ``(new_box, resume_at)``. ``resume_at`` is ``None`` if all of the
    content fits. Otherwise it can be passed as a ``skip_stack`` parameter
    to resume where we left off.

    ``new_box`` is non-empty (unless the box is empty) and as big as possible
    while being narrower than ``available_width``, if possible (may overflow
    is no split is possible.)

    """
    resolve_percentages(box, containing_block)
    if isinstance(box, boxes.TextBox):
        box.position_x = position_x
        if skip_stack is None:
            skip = 0
        else:
            skip, skip_stack = skip_stack
            skip = skip or 0
            assert skip_stack is None

        new_box, skip, preserved_line_break = split_text_box(
            document, box, max_x - position_x, skip)

        if skip is None:
            resume_at = None
        else:
            resume_at = (skip, None)
    elif isinstance(box, boxes.InlineBox):
        if box.margin_left == 'auto':
            box.margin_left = 0
        if box.margin_right == 'auto':
            box.margin_right = 0
        new_box, resume_at, preserved_line_break = split_inline_box(
            document, box, position_x, max_x, skip_stack, containing_block,
            device_size)
    elif isinstance(box, boxes.AtomicInlineLevelBox):
        new_box = atomic_box(box, containing_block, device_size)
        new_box.position_x = position_x
        new_box.baseline = new_box.margin_height()
        resume_at = None
        preserved_line_break = False
    #else: unexpected box type here
    return new_box, resume_at, preserved_line_break


def split_inline_box(document, box, position_x, max_x, skip_stack,
                     containing_block, device_size):
    """Same behavior as split_inline_level."""
    initial_position_x = position_x
    assert isinstance(box, (boxes.LineBox, boxes.InlineBox))
    left_spacing = (box.padding_left + box.margin_left +
                    box.style.border_left_width)
    right_spacing = (box.padding_right + box.margin_right +
                     box.style.border_right_width)
    position_x += left_spacing
    content_box_left = position_x

    children = []
    preserved_line_break = False

    if skip_stack is None:
        skip = 0
    else:
        skip, skip_stack = skip_stack

    for index, child in box.enumerate_skip(skip):
        assert child.is_in_normal_flow(), '"Abnormal" flow not supported yet.'
        new_child, resume_at, preserved = split_inline_level(
            document, child, position_x, max_x, skip_stack,
            containing_block, device_size)
        skip_stack = None
        if preserved:
            preserved_line_break = True

        # TODO: this is non-optimal when last_child is True and
        #   width <= remaining_width < width + right_spacing
        # with
        #   width = part1.margin_width()

        # TODO: on the last child, take care of right_spacing

        if new_child is None:
            # may be None where we would have an empty TextBox
            assert isinstance(child, boxes.TextBox)
        else:
            margin_width = new_child.margin_width()
            new_position_x = position_x + margin_width

            if (new_position_x > max_x and children):
                # too wide, and the inline is non-empty:
                # put child entirely on the next line.
                resume_at = (index, None)
                break
            else:
                position_x = new_position_x
                children.append(new_child)

        if resume_at is not None:
            resume_at = (index, resume_at)
            break
    else:
        resume_at = None

    new_box = box.copy_with_children(children)
    new_box.position_x = initial_position_x
    new_box.width = position_x - content_box_left

    # Create a "strut":
    # http://www.w3.org/TR/CSS21/visudet.html#strut
    # TODO: cache these results for a given set of styles?
    fragment = TextFragment(
        b'', box.style, cairo.Context(document.surface))
    _, _, _, height, baseline, _ = fragment.split_first_line()
    leading = box.style.line_height - height
    half_leading = leading / 2.
    # Set margins to the half leading but also compensate for borders and
    # paddings. We want margin_height() == line_height
    new_box.margin_top = (half_leading - new_box.style.border_top_width -
                          new_box.padding_bottom)
    new_box.margin_bottom = (half_leading - new_box.style.border_bottom_width -
                             new_box.padding_bottom)
    # form the top of the content box
    new_box.baseline = baseline
    # form the top of the margin box
    new_box.baseline += half_leading
    new_box.height = height

    if resume_at is not None:
        # There is a line break inside this box.
        box.reset_spacing('left')
        new_box.reset_spacing('right')
    return new_box, resume_at, preserved_line_break


def split_text_box(document, box, available_width, skip):
    """Keep as much text as possible from a TextBox in a limitied width.
    Try not to overflow but always have some text in ``new_box``

    Return ``(new_box, skip)``. ``skip`` is the number of UTF-8 bytes
    to skip form the start of the TextBox for the next line, or ``None``
    if all of the text fits.

    Also break an preserved whitespace.

    """
    assert isinstance(box, boxes.TextBox)
    font_size = box.style.font_size
    text = box.text[skip:]
    if font_size == 0 or available_width <= 0 or not text:
        return None, None, False

    fragment = TextFragment(text, box.style,
        cairo.Context(document.surface), available_width)

    # XXX ``resume_at`` is an index in UTF-8 bytes, not unicode codepoints.
    show_line, length, width, height, baseline, resume_at = \
        fragment.split_first_line()

    # Convert ``length`` and ``resume_at`` from UTF-8 indexes in text
    # to Unicode indexes.
    # No need to encode what’s after resume_at (if set) or length (if
    # resume_at is not set). One code point is one or more byte, so
    # UTF-8 indexes are always bigger or equal to Unicode indexes.
    partial_text = text[:resume_at or length]
    utf8_text = partial_text.encode('utf8')
    new_text = utf8_text[:length].decode('utf8')
    new_length = len(new_text)
    if resume_at is not None:
        between = utf8_text[length:resume_at].decode('utf8')
        resume_at = new_length + len(between)
    length = new_length

    if length > 0:
        box = box.copy_with_text(new_text)
        box.width = width
        box.show_line = show_line
        # "The height of the content area should be based on the font,
        #  but this specification does not specify how."
        # http://www.w3.org/TR/CSS21/visudet.html#inline-non-replaced
        # We trust Pango and use the height of the LayoutLine.
        # It is based on font_size (slightly larger), but I’m not sure how.
        # TODO: investigate this
        box.height = height
        # "only the 'line-height' is used when calculating the height
        #  of the line box."
        # Set margins so that margin_height() == line_height
        leading = box.style.line_height - height
        half_leading = leading / 2.
        box.margin_top = half_leading
        box.margin_bottom = half_leading
        # form the top of the content box
        box.baseline = baseline
        # form the top of the margin box
        box.baseline += (box.margin_top + box.style.border_top_width +
                         box.padding_top)
    else:
        box = None

    if resume_at is None:
        preserved_line_break = False
    else:
        preserved_line_break = (length != resume_at)
        if preserved_line_break:
            assert between == '\n', ('Got %r between two lines. '
                'Expected nothing or a preserved line break' % (between,))
        resume_at += skip

    return box, resume_at, preserved_line_break


def inline_box_verticality(box, baseline_y):
    """Handle ``vertical-align`` within an :class:`InlineBox`.

    Place all boxes vertically assuming that the baseline of ``box``
    is at `y = baseline_y`.

    Return ``(max_y, min_y)``, the maximum and minimum vertical position
    of margin boxes.

    """
    max_y = None
    min_y = None
    for child in box.children:
        vertical_align = child.style.vertical_align
        if vertical_align == 'baseline':
            child_baseline_y = baseline_y
        elif vertical_align == 'middle':
            # TODO: find ex from font metrics
            one_ex = box.style.font_size * 0.5
            top = baseline_y - (one_ex + child.margin_height()) / 2.
            child_baseline_y = top + child.baseline
        # TODO: actually implement vertical-align: top and bottom
        elif vertical_align in ('text-top', 'top'):
            # align top with the top of the parent’s content area
            top = (baseline_y - box.baseline + box.margin_top +
                   box.style.border_top_width + box.padding_top)
            child_baseline_y = top + child.baseline
        elif vertical_align in ('text-bottom', 'bottom'):
            # align bottom with the bottom of the parent’s content area
            bottom = (baseline_y - box.baseline + box.margin_top +
                      box.style.border_top_width + box.padding_top +
                      box.height)
            child_baseline_y = bottom - child.margin_height() + child.baseline
        else:
            # Numeric value: The child’s baseline is `vertical_align` above
            # (lower y) the parent’s baseline.
            child_baseline_y = baseline_y - vertical_align
        # the child’s `top` is `child.baseline` above (lower y) its baseline.
        top = child_baseline_y - child.baseline
        child.position_y = top
        bottom = top + child.margin_height()
        if min_y is None or top < min_y:
            min_y = top
        if max_y is None or bottom > max_y:
            max_y = bottom
        if isinstance(child, boxes.InlineBox):
            children_max_y, children_min_y = inline_box_verticality(
                child, child_baseline_y)
            if children_max_y is None:
                if (
                    child.margin_width() == 0
                    # Guard against the case where a negative margin
                    # compensates something else.
                    and child.margin_left == 0
                    and child.margin_right == 0
                ):
                    # No content, ignore this box’s line-height.
                    # See http://www.w3.org/TR/CSS21/visuren.html#phantom-line-box
                    child.position_y = child_baseline_y
                    child.height = 0
                    continue
            else:
                assert children_min_y is not None
                if children_min_y < min_y:
                    min_y = children_min_y
                if children_max_y > max_y:
                    max_y = children_max_y
    return max_y, min_y


def text_align(line, containing_block):
    """Return how much the line should be moved horizontally according to
    the `text-align` property.

    """
    align = line.style.text_align
    if align == 'left':
        return 0
    offset = containing_block.width - line.width
    if align == 'center':
        offset /= 2.
    else:
        assert align == 'right'
    return offset



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
WeasyPrint
==========

WeasyPrint converts web documents, mainly HTML documents with CSS, to PDF.

"""

VERSION = __version__ = '0.3.1'  # Also change this in setup.py


import argparse


__all__ = ['main']


FORMATS = {
    'pdf': 'PDFDocument',
    'png': 'PNGDocument',
}


# cssutils defaults to logging to stderr, but we want to hide its validation
# warnings as we’re doing our own validation.
#logging.getLogger('CSSUTILS').handlers[:] = []


def _join(sequence, key=lambda x: x):
    """Return a string of the sorted elements of ``sequence``.

    The two last elements are separated by ' or ', the other ones are separated
    by ', '.

    If a ``key`` function is given, this function is applied to the elements of
    ``sequence`` before joining them.

    """
    sequence = sorted(sequence)
    last = key(sequence[-1])
    if len(sequence) == 1:
        return last
    else:
        return ' or '.join([', '.join(map(key, sequence[:-1])), last])


def main():
    """Parse command-line arguments and convert the given document."""
    extensions = _join(FORMATS, lambda x: '.' + x)

    parser = argparse.ArgumentParser(
        description='Renders web pages into ' + _join(FORMATS, str.upper))
    parser.add_argument('--version', action='version',
                        version='WeasyPrint version %s' % VERSION,
                        help='Print WeasyPrint’s version number and exit.')
    parser.add_argument('-e', '--encoding',
                        help='Character encoding of the input')
    parser.add_argument('-f', '--format', choices=FORMATS,
                        help='Output format. Can be ommited if `output` '
                             'ends with ' + extensions)
    parser.add_argument('-s', '--stylesheet', action='append',
                        help='Apply a user stylesheet to the document. '
                             'May be given multiple times.')
    parser.add_argument('input',
        help='URL or filename of the HTML input, or - for stdin')
    parser.add_argument('output',
        help='Filename where output is written, or - for stdout')

    args = parser.parse_args()

    # Import here so that these are not needed for --version and --help
    import sys
    import logging
    import cssutils

    from . import document
    from .utils import ensure_url

    if args.format is None:
        for file_format in FORMATS:
            if args.output.endswith('.' + file_format):
                args.format = file_format
                break
        else:
            parser.error(
                'Either sepecify a format with -f or choose an '
                'output filename that ends in ' + extensions)

    if args.input == '-':
        args.input = sys.stdin

    if args.output == '-':
        args.output = sys.stdout

    if hasattr(logging, 'NullHandler'):
        # New in 2.7
        cssutils_logger = logging.getLogger('CSSUTILS')
        del cssutils_logger.handlers[:]
        cssutils_logger.addHandler(logging.NullHandler())

    logger = logging.getLogger('WEASYPRINT')
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler())

    users_stylesheets = [
        cssutils.parseUrl(ensure_url(filename_or_url))
        for filename_or_url in args.stylesheet or []]

    document_class = getattr(document, FORMATS[args.format])
    doc = document_class.from_file(args.input, encoding=args.encoding,
        user_stylesheets=users_stylesheets)
    doc.write_to(args.output)



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Functions laying out the block boxes.

"""

from __future__ import division

from .inlines import get_next_linebox, replaced_box_width, replaced_box_height
from .markers import list_marker_layout
from .tables import table_layout, fixed_table_layout
from .percentages import resolve_percentages
from ..formatting_structure import boxes


def block_level_layout(document, box, max_position_y, skip_stack,
                       containing_block, device_size, page_is_empty):
    """Lay out the block-level ``box``.

    :param max_position_y: the absolute vertical position (as in
                           ``some_box.position_y``) of the bottom of the
                           content box of the current page area.

    """
    if isinstance(box, boxes.TableBox):
        return table_layout(document, box, max_position_y, containing_block,
                            device_size, page_is_empty)
    elif isinstance(box, boxes.BlockBox):
        if box.is_table_wrapper:
            return block_table_wrapper(document, box, max_position_y,
                skip_stack, containing_block, device_size, page_is_empty)
        else:
            return block_box_layout(document, box, max_position_y, skip_stack,
                containing_block, device_size, page_is_empty)
    elif isinstance(box, boxes.BlockReplacedBox):
        return block_replaced_box_layout(
            box, containing_block, device_size), None
    else:
        raise TypeError('Layout for %s not handled yet' % type(box).__name__)


def block_box_layout(document, box, max_position_y, skip_stack,
                     containing_block, device_size, page_is_empty):
    """Lay out the block ``box``."""
    resolve_percentages(box, containing_block)
    block_level_width(box, containing_block)
    list_marker_layout(document, box, containing_block)
    return block_level_height(document, box, max_position_y, skip_stack,
        device_size, page_is_empty)


def block_replaced_box_layout(box, containing_block, device_size):
    """Lay out the block :class:`boxes.ReplacedBox` ``box``."""
    assert isinstance(box, boxes.ReplacedBox)
    resolve_percentages(box, containing_block)

    # http://www.w3.org/TR/CSS21/visudet.html#block-replaced-width
    replaced_box_width(box, device_size)
    block_level_width(box, containing_block)

    # http://www.w3.org/TR/CSS21/visudet.html#inline-replaced-height
    replaced_box_height(box, device_size)
    if box.margin_top == 'auto':
        box.margin_top = 0
    if box.margin_bottom == 'auto':
        box.margin_bottom = 0

    return box


def block_level_width(box, containing_block):
    """Set the ``box`` width."""
    # 'cb' stands for 'containing block'
    cb_width = containing_block.width

    # http://www.w3.org/TR/CSS21/visudet.html#blockwidth

    # These names are waaay too long
    margin_l = box.margin_left
    margin_r = box.margin_right
    padding_l = box.padding_left
    padding_r = box.padding_right
    border_l = box.style.border_left_width
    border_r = box.style.border_right_width
    width = box.width

    # Only margin-left, margin-right and width can be 'auto'.
    # We want:  width of containing block ==
    #               margin-left + border-left-width + padding-left + width
    #               + padding-right + border-right-width + margin-right

    paddings_plus_borders = padding_l + padding_r + border_l + border_r
    if box.width != 'auto':
        total = paddings_plus_borders + width
        if margin_l != 'auto':
            total += margin_l
        if margin_r != 'auto':
            total += margin_r
        if total > cb_width:
            if margin_l == 'auto':
                margin_l = box.margin_left = 0
            if margin_r == 'auto':
                margin_r = box.margin_right = 0
    if width != 'auto' and margin_l != 'auto' and margin_r != 'auto':
        # The equation is over-constrained
        margin_sum = cb_width - paddings_plus_borders - width
        if containing_block.direction == 'ltr':
            margin_r = box.margin_right = margin_sum - margin_l
        else:
            margin_l = box.margin_left = margin_sum - margin_r
    if width == 'auto':
        if margin_l == 'auto':
            margin_l = box.margin_left = 0
        if margin_r == 'auto':
            margin_r = box.margin_right = 0
        width = box.width = cb_width - (
            paddings_plus_borders + margin_l + margin_r)
    margin_sum = cb_width - paddings_plus_borders - width
    if margin_l == 'auto' and margin_r == 'auto':
        box.margin_left = margin_sum / 2.
        box.margin_right = margin_sum / 2.
    elif margin_l == 'auto' and margin_r != 'auto':
        box.margin_left = margin_sum - margin_r
    elif margin_l != 'auto' and margin_r == 'auto':
        box.margin_right = margin_sum - margin_l


def block_level_height(document, box, max_position_y, skip_stack,
                       device_size, page_is_empty):
    """Set the ``box`` height."""
    assert isinstance(box, boxes.BlockContainerBox)

    if box.style.overflow != 'visible':
        raise NotImplementedError

    if box.margin_top == 'auto':
        box.margin_top = 0
    if box.margin_bottom == 'auto':
        box.margin_bottom = 0

    position_x = box.content_box_x()
    position_y = box.content_box_y()
    initial_position_y = position_y

    new_children = []

    if skip_stack is None:
        skip = 0
    else:
        skip, skip_stack = skip_stack

    for index, child in box.enumerate_skip(skip):
        if not child.is_in_normal_flow():
            continue
        # TODO: collapse margins
        # See http://www.w3.org/TR/CSS21/visudet.html#normal-block
        child.position_x = position_x
        child.position_y = position_y
        if isinstance(child, boxes.LineBox):
            assert len(box.children) == 1, (
                'line box with siblings before layout')
            is_page_break = False
            while 1:
                new_containing_block = box
                line, resume_at = get_next_linebox(
                    document, child, position_y, skip_stack,
                    new_containing_block, device_size)
                if line is None:
                    break
                new_position_y = position_y + line.height
                # Allow overflow if the first line of the page is higher
                # than the page itself so that we put *something* on this
                # page and can advance in the document.
                if new_position_y > max_position_y and not page_is_empty:
                    if not new_children:
                        # Page break before any content, cancel the whole box.
                        return None, None
                    # Page break here, resume before this line
                    resume_at = (index, skip_stack)
                    is_page_break = True
                    break
                new_children.append(line)
                page_is_empty = False
                position_y = new_position_y
                if resume_at is None:
                    break
                skip_stack = resume_at
            if is_page_break:
                break
        else:
            new_containing_block = box
            new_child, resume_at = block_level_layout(
                document, child, max_position_y, skip_stack,
                new_containing_block, device_size, page_is_empty)
            skip_stack = None
            if new_child is None:
                if new_children:
                    resume_at = (index, None)
                    break
                else:
                    # This was the first child of this box, cancel the box
                    # completly
                    return None, None
            new_position_y = position_y + new_child.margin_height()
            # Bottom borders may overflow here
            # TODO: back-track somehow when all lines fit but not borders
            new_children.append(new_child)
            page_is_empty = False
            position_y = new_position_y
            if resume_at is not None:
                resume_at = (index, resume_at)
                break
    else:
        resume_at = None

    new_box = box.copy_with_children(new_children)

    if new_box.height == 'auto':
        new_box.height = position_y - initial_position_y

    if resume_at is not None:
        # If there was a list marker, we kept it on `new_box`.
        # Do not repeat it on `box` on the next page.
        # TODO: Do this non-destructively
        box.outside_list_marker = None
        box.reset_spacing('top')
        new_box.reset_spacing('bottom')
    return new_box, resume_at


def block_table_wrapper(document, wrapper, max_position_y, skip_stack,
                        containing_block, device_size, page_is_empty):
    """Layout for the wrapper of a block-level table wrapper."""
    for child in wrapper.children:
        if isinstance(child, boxes.TableBox):
            table = child
            break
    else:
        raise ValueError('Table wrapper without a table')
    resolve_percentages(wrapper, containing_block)
    resolve_percentages(table, containing_block)
    # Count the wrapper margins in case of `width: auto`
    table.margin_left = wrapper.margin_left
    table.margin_right = wrapper.margin_right
    block_level_width(table, containing_block)
    # The table margins are on the table wrapper box, not on the table box
    table.margin_left = 0
    table.margin_right = 0

    fixed_table_layout(table)
    wrapper.width = wrapper.style.width = table.border_width()
    return block_box_layout(document, wrapper, max_position_y, skip_stack,
                            containing_block, device_size, page_is_empty)



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Module managing the layout creation before drawing a document.

"""

from __future__ import division

from .. import css
from .blocks import block_box_layout
from .percentages import resolve_percentages
from ..formatting_structure import boxes


def make_page(document, page_number, resume_at):
    """Take just enough content from the beginning to fill one page.

    Return ``page, finished``. ``page`` is a laid out Page object, ``finished``
    is ``True`` if there is no more content, this was the last page.

    """
    style = css.page_style(document, page_number)
    page = boxes.PageBox(page_number, style)

    device_size = page.style.size
    page.outer_width, page.outer_height = device_size

    sheet = lambda: 1  # dummy object holding attributes.
    sheet.width, sheet.height = device_size
    resolve_percentages(page, sheet)

    page.position_x = 0
    page.position_y = 0
    page.width = page.outer_width - page.horizontal_surroundings()
    page.height = page.outer_height - page.vertical_surroundings()

    root_box = document.formatting_structure

    # This box is not laid-out yet, but it may be needed during layout
    # eg. for ``page.direction``.
    page.root_box = root_box

    root_box.position_x = page.content_box_x()
    root_box.position_y = page.content_box_y()
    page_content_bottom = root_box.position_y + page.height
    initial_containing_block = page

    # TODO: handle cases where the root element is something else.
    # See http://www.w3.org/TR/CSS21/visuren.html#dis-pos-flo
    assert isinstance(root_box, boxes.BlockBox)
    page.root_box, resume_at = block_box_layout(
        document, root_box, page_content_bottom, resume_at,
        initial_containing_block, device_size, page_is_empty=True)
    assert page.root_box

    return page, resume_at


def layout(document):
    """Lay out the whole document.

    This includes line breaks, page breaks, absolute size and position for all
    boxes.

    :param document: a Document object.
    :returns: a list of laid out Page objects.

    """
    pages = []
    page_number = 1
    resume_at = None
    while True:
        page, resume_at = make_page(document, page_number, resume_at)
        pages.append(page)
        if resume_at is None:
            return pages
        page_number += 1



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Handle various image formats.
"""

from __future__ import division
from StringIO import StringIO

import cairo

from .utils import urlopen


# Map MIME types to functions that take a byte stream and return
# ``(surface, width, height)`` a cairo Surface and its dimension in pixels.
FORMAT_HANDLERS = {}

# TODO: currently CairoSVG only support images with an explicit
# width and height. When it supports images with only an intrinsic ratio
# this API will need to change.


def register_format(mime_type):
    """Register a handler for a give image MIME type."""
    def decorator(function):
        FORMAT_HANDLERS[mime_type] = function
        return function
    return decorator


@register_format('image/png')
def png_handler(file_like):
    """Return a cairo Surface from a PNG byte stream."""
    surface = cairo.ImageSurface.create_from_png(file_like)
    return surface, surface.get_width(), surface.get_height()


@register_format('image/svg+xml')
def cairosvg_handler(file_like):
    """Return a cairo Surface from a SVG byte stream.

    This handler uses CairoSVG: http://cairosvg.org/
    """
    # TODO: also offer librsvg?
    try:
        import cairosvg
    except ImportError:
        return None
    from cairosvg.surface_type import DummySurface
    # TODO: find a way to pass file_like to the parser, not read the
    # whole string in memory.
    surface = cairosvg.svg2surface(file_like, DummySurface)
    return surface.cairo, surface.width, surface.height


def fallback_handler(file_like):
    """
    Parse a byte stream with PIL and return a cairo Surface.

    PIL supports many raster image formats and does not take a `format`
    parameter, it guesses the format from the content.
    """
    try:
        from PIL import Image
    except ImportError:
        try:
            # It is sometimes installed with another name...
            import Image
        except ImportError:
            return None  # PIL is not installed
    if not (hasattr(file_like, 'seek') and hasattr(file_like, 'tell')):
        # PIL likes to have these methods
        file_like = StringIO(file_like.read())
    png = StringIO()
    image = Image.open(file_like)
    image = image.convert('RGBA')
    image.save(png, "PNG")
    png.seek(0)
    return png_handler(png)


def get_image_surface_from_uri(uri):
    """Get a :class:`cairo.Surface`` from an image URI."""
    try:
        file_like, mime_type, _charset = urlopen(uri)
    except IOError:
        # TODO: warn
        return None
    # TODO: implement image type sniffing?
# http://www.w3.org/TR/html5/fetching-resources.html#content-type-sniffing:-image
    handler = FORMAT_HANDLERS.get(mime_type, fallback_handler)
    try:
        return handler(file_like)
    except (IOError, MemoryError, NotImplementedError):
        # Network or parsing error
        # TODO: warn
        return None
    finally:
        file_like.close()
